# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Finetuning the library models for sequence classification on GLUE."""


import dataclasses
import logging
import os
import sys
from dataclasses import dataclass, field
from typing import Callable, Dict, Optional

import torch.utils.data
import torch.nn.functional as F
import numpy as np
from scipy.special import softmax
import calibration as cal
from models import TempCNN, TempConst, TempResNet, TempLinearOnReps, TempNNOnReps, CalibrationNNOnReps
import utils

from transformers import AutoConfig, AutoModelForSequenceClassification, AutoTokenizer, EvalPrediction, GlueDataset
from transformers import GlueDataTrainingArguments as DataTrainingArguments
from transformers import (
    HfArgumentParser,
    Trainer,
    TrainingArguments,
    glue_compute_metrics,
    glue_output_modes,
    glue_tasks_num_labels,
    set_seed,
)
from calib_bert_trainer import CalibrationTrainer


logger = logging.getLogger(__name__)


@dataclass
class ModelArguments:
    """
    Arguments pertaining to which model/config/tokenizer we are going to fine-tune from.
    """

    model_name_or_path: str = field(
        metadata={"help": "Path to pretrained model or model identifier from huggingface.co/models"}
    )
    config_name: Optional[str] = field(
        default=None, metadata={"help": "Pretrained config name or path if not the same as model_name"}
    )
    tokenizer_name: Optional[str] = field(
        default=None, metadata={"help": "Pretrained tokenizer name or path if not the same as model_name"}
    )
    cache_dir: Optional[str] = field(
        default=None, metadata={"help": "Where do you want to store the pretrained models downloaded from s3"}
    )

@dataclass
class CustomArguments:
    """
    Customized arguments.
    """
    split_size: Optional[int] = field(
        default=-1, metadata={"help": "Split training data into _ and (1-_) proportions"}
    )
    use_split: Optional[str] = field(
        default='train', metadata={"help": "which split to train on: {train|train_val}"}
    )
    use_temp: Optional[bool] = field(
        default=False, metadata={"help": "use temperature scaling models"}
    )
    use_calib: Optional[bool] = field(
        default=False, metadata={"help": "use generalized calibration models"}
    )
    temp_model: Optional[str] = field(
        default='const', metadata={"help": "type of temperature model"}
    )
    min_temp: Optional[float] = field(
        default=0.2, metadata={"help": "minimal temperature in temperature model"}
    )
    temp_model_depth: Optional[int] = field(
        default=2, metadata={"help": "depth of temperature model"}
    )
    temp_model_width: Optional[int] = field(
        default=512, metadata={"help": "width of temperature model"}
    )
    temp_init_increment: Optional[float] = field(
        default=0.5, metadata={"help": "increment for initializing individual temps"}
    )
    temp_model_on_bert_layer: Optional[int] = field(
        default=12, metadata={"help": "temp model attached on _th BERT layer"}
    )
    num_temps: Optional[int] = field(
        default=5, metadata={"help": "number of temperature models"}
    )
    neg_slope: Optional[float] = field(
        default=0.5, metadata={"help": "negative slope for leaky relu in calibration model"}
    )
    calib_output_mode: Optional[str] = field(
        default='logprobs', metadata={"help": "output mode for calibration model"}
    )
    nll_weight: Optional[float] = field(
        default=1.0, metadata={"help": "weight of nll loss in post calibration"}
    )
    ece_weight: Optional[float] = field(
        default=0.0, metadata={"help": "weight of ece loss in post calibration"}
    )
    ece_adaptive_binning: Optional[bool] = field(
        default=0.0, metadata={"help": "use adaptive binning in ece training"}
    )
    ece_num_partitions_train: Optional[int] = field(
        default=5, metadata={"help": "number of bins in ece loss at train time"}
    )
    ece_num_partitions_val: Optional[int] = field(
        default=15, metadata={"help": "number of bins in ece loss at validation time"}
    )
    optimizer: Optional[str] = field(
        default='sgd', metadata={"help": "optimizer"}
    )
    scheduler: Optional[str] = field(
        default='onecycle', metadata={"help": "learning rate scheduler"}
    )

def main():
    # See all possible arguments in src/transformers/training_args.py
    # or by passing the --help flag to this script.
    # We now keep distinct sets of args, for a cleaner separation of concerns.

    parser = HfArgumentParser((ModelArguments, DataTrainingArguments, TrainingArguments, CustomArguments))

    if len(sys.argv) == 2 and sys.argv[1].endswith(".json"):
        # If we pass only one argument to the script and it's the path to a json file,
        # let's parse it to get our arguments.
        model_args, data_args, training_args, custom_args = parser.parse_json_file(json_file=os.path.abspath(sys.argv[1]))
    else:
        model_args, data_args, training_args, custom_args = parser.parse_args_into_dataclasses()

    if (
        os.path.exists(training_args.output_dir)
        and os.listdir(training_args.output_dir)
        and training_args.do_train
        and not training_args.overwrite_output_dir
    ):
        raise ValueError(
            f"Output directory ({training_args.output_dir}) already exists and is not empty. Use --overwrite_output_dir to overcome."
        )

    # Default tensorboard logging dir to model output dir
    training_args.logging_dir = training_args.output_dir

    # Setup logging
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s -   %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        level=logging.INFO if training_args.local_rank in [-1, 0] else logging.WARN,
    )
    logger.warning(
        "Process rank: %s, device: %s, n_gpu: %s, distributed training: %s, 16-bits training: %s",
        training_args.local_rank,
        training_args.device,
        training_args.n_gpu,
        bool(training_args.local_rank != -1),
        training_args.fp16,
    )
    logger.info("Training/evaluation parameters %s", training_args)

    # Set seed
    set_seed(training_args.seed)

    try:
        num_labels = glue_tasks_num_labels[data_args.task_name]
        output_mode = glue_output_modes[data_args.task_name]
    except KeyError:
        raise ValueError("Task not found: %s" % (data_args.task_name))

    # Load pretrained model and tokenizer
    #
    # Distributed training:
    # The .from_pretrained methods guarantee that only one local process can concurrently
    # download model & vocab.

    config = AutoConfig.from_pretrained(
        model_args.config_name if model_args.config_name else model_args.model_name_or_path,
        num_labels=num_labels,
        finetuning_task=data_args.task_name,
        cache_dir=model_args.cache_dir,
    )
    tokenizer = AutoTokenizer.from_pretrained(
        model_args.tokenizer_name if model_args.tokenizer_name else model_args.model_name_or_path,
        cache_dir=model_args.cache_dir,
    )
    model = AutoModelForSequenceClassification.from_pretrained(
        model_args.model_name_or_path,
        from_tf=bool(".ckpt" in model_args.model_name_or_path),
        config=config,
        cache_dir=model_args.cache_dir,
    )

    # Get datasets
    train_dataset = (
        GlueDataset(data_args, tokenizer=tokenizer, cache_dir=model_args.cache_dir) if training_args.do_train else None
    )
    eval_dataset = (
        GlueDataset(data_args, tokenizer=tokenizer, mode="dev", cache_dir=model_args.cache_dir)
        if training_args.do_eval
        else None
    )
    test_dataset = (
        GlueDataset(data_args, tokenizer=tokenizer, mode="test", cache_dir=model_args.cache_dir)
        if training_args.do_predict
        else None
    )
    train_val_dataset = None
    if training_args.do_train and custom_args.split_size > 0:
        train_splits = torch.utils.data.random_split(
            train_dataset, [custom_args.split_size, len(train_dataset)-custom_args.split_size],
            generator=torch.Generator().manual_seed(42)
        )
        train_dataset, train_val_dataset = train_splits[0], train_splits[1]
    if custom_args.use_split == 'train':
        dataset = train_dataset
    else:
        dataset = train_val_dataset

    def build_compute_metrics_fn(task_name: str) -> Callable[[EvalPrediction], Dict]:
        def compute_metrics_fn(p: EvalPrediction):
            if output_mode == "classification":
                preds = np.argmax(p.predictions, axis=1)
            elif output_mode == "regression":
                preds = np.squeeze(p.predictions)
            results = glue_compute_metrics(task_name, preds, p.label_ids)
            probs, targets = F.softmax(torch.from_numpy(p.predictions)), torch.from_numpy(p.label_ids)
            ece = utils.expected_calibration_error(probs, targets,
                                                   num_partitions=custom_args.ece_num_partitions_val,
                                                   verbose=False)
            debiased_ece_l1 = cal.get_calibration_error(probs, targets,
                                                        p=1, debias=True, mode='top-label')
            debiased_ece_l2 = cal.get_calibration_error(probs, targets,
                                                        p=2, debias=True, mode='top-label')
            results['ece'] = ece.item()
            results['debiased_ece_l1'] = debiased_ece_l1
            results['debiased_ece_l2'] = debiased_ece_l2
            results['pred_ent'] = utils.predictive_entropy(probs).item()
            return results

        return compute_metrics_fn

    # Define post-calibration models
    temp_model, calib_model = None, None
    if custom_args.use_temp or custom_args.use_calib:
        if custom_args.use_temp:
            if custom_args.temp_model == 'const':
                temp_model = TempConst(min_temp=custom_args.min_temp).to(training_args.device)
            elif custom_args.temp_model == 'nn_on_reps':
                temp_model = TempNNOnReps([custom_args.temp_model_on_bert_layer], [768],
                                          depth=custom_args.temp_model_depth, width=custom_args.temp_model_width,
                                          min_temp=custom_args.min_temp).to(training_args.device)
            the_model = temp_model
        elif custom_args.use_calib:
            if custom_args.temp_model == 'nn_on_reps':
                calib_model = CalibrationNNOnReps(
                    [custom_args.temp_model_on_bert_layer], [768],
                    depth=custom_args.temp_model_depth, width=custom_args.temp_model_width,
                    min_temp=custom_args.min_temp, temp_init_increment=custom_args.temp_init_increment,
                    num_temps=custom_args.num_temps, neg_slope=custom_args.neg_slope,
                    output_mode=custom_args.calib_output_mode).to(training_args.device)
            the_model = calib_model
        if custom_args.optimizer == 'sgd':
            optimizer = torch.optim.SGD(the_model.parameters(), training_args.learning_rate,
                                        weight_decay=training_args.weight_decay)
        if custom_args.scheduler == 'onecycle':
            scheduler = torch.optim.lr_scheduler.OneCycleLR(
                optimizer, max_lr=training_args.learning_rate,
                total_steps=int(np.ceil(len(dataset) / training_args.train_batch_size) * training_args.num_train_epochs)
            )

    # Initialize our Trainer
    if custom_args.use_temp or custom_args.use_calib:
        trainer = CalibrationTrainer(
            model=model,
            temp_model=temp_model,
            calib_model=calib_model,
            temp_model_type=custom_args.temp_model,
            optimizers=(optimizer, scheduler),
            args=training_args,
            train_dataset=dataset,
            eval_dataset=eval_dataset,
            compute_metrics=build_compute_metrics_fn(data_args.task_name),
            nll_weight=custom_args.nll_weight,
            ece_weight=custom_args.ece_weight,
            ece_num_partitions_train=custom_args.ece_num_partitions_train,
        )
    else:
        trainer = Trainer(
            model=model,
            args=training_args,
            train_dataset=dataset,
            eval_dataset=eval_dataset,
            compute_metrics=build_compute_metrics_fn(data_args.task_name),
        )

    # Training
    if training_args.do_train:
        trainer.train(
            model_path=model_args.model_name_or_path if os.path.isdir(model_args.model_name_or_path) else None
        )
        trainer.save_model()
        # For convenience, we also re-save the tokenizer to the same directory,
        # so that you can share your model easily on huggingface.co/models =)
        if trainer.is_world_master():
            tokenizer.save_pretrained(training_args.output_dir)

    # Evaluation
    eval_results = {}
    if training_args.do_eval:
        logger.info("*** Evaluate ***")

        # Loop to handle MNLI double evaluation (matched, mis-matched)
        eval_datasets = [eval_dataset]
        if data_args.task_name == "mnli":
            mnli_mm_data_args = dataclasses.replace(data_args, task_name="mnli-mm")
            eval_datasets.append(
                GlueDataset(mnli_mm_data_args, tokenizer=tokenizer, mode="dev", cache_dir=model_args.cache_dir)
            )

        for eval_dataset in eval_datasets:
            trainer.compute_metrics = build_compute_metrics_fn(eval_dataset.args.task_name)
            eval_result = trainer.evaluate(eval_dataset=eval_dataset)

            output_eval_file = os.path.join(
                training_args.output_dir, f"eval_results_{eval_dataset.args.task_name}.txt"
            )
            if trainer.is_world_master():
                with open(output_eval_file, "w") as writer:
                    logger.info("***** Eval results {} *****".format(eval_dataset.args.task_name))
                    for key, value in eval_result.items():
                        logger.info("  %s = %s", key, value)
                        writer.write("%s = %s\n" % (key, value))

            eval_results.update(eval_result)

    if training_args.do_predict:
        logging.info("*** Test ***")
        test_datasets = [test_dataset]
        if data_args.task_name == "mnli":
            mnli_mm_data_args = dataclasses.replace(data_args, task_name="mnli-mm")
            test_datasets.append(
                GlueDataset(mnli_mm_data_args, tokenizer=tokenizer, mode="test", cache_dir=model_args.cache_dir)
            )

        for test_dataset in test_datasets:
            predictions = trainer.predict(test_dataset=test_dataset).predictions
            if output_mode == "classification":
                predictions = np.argmax(predictions, axis=1)

            output_test_file = os.path.join(
                training_args.output_dir, f"test_results_{test_dataset.args.task_name}.txt"
            )
            if trainer.is_world_master():
                with open(output_test_file, "w") as writer:
                    logger.info("***** Test results {} *****".format(test_dataset.args.task_name))
                    writer.write("index\tprediction\n")
                    for index, item in enumerate(predictions):
                        if output_mode == "regression":
                            writer.write("%d\t%3.3f\n" % (index, item))
                        else:
                            item = test_dataset.get_labels()[item]
                            writer.write("%d\t%s\n" % (index, item))
    return eval_results


def _mp_fn(index):
    # For xla_spawn (TPUs)
    main()


if __name__ == "__main__":
    main()
